import warnings

from chainer.utils import argument

from chainermn.communicators.communicator_base import CommunicatorBase  # NOQA


def create_communicator(
        communicator_name='pure_nccl', mpi_comm=None, **kwargs):
    """Create a ChainerMN communicator.

    Different communicators provide different approaches of communication, so
    they have different performance charasteristics. The default communicator
    ``pure_nccl`` is expected to generally perform well on a variety of
    environments, so one need not to change communicators in most cases.
    However, you may need to choose other communicators depending on
    your computing platform and the availability of NCCL library.
    The following communicators are available.

    +---------------+---+---+--------+--------------------------------------+
    |Name           |CPU|GPU|NCCL    |Recommended Use Cases                 |
    +===============+===+===+========+======================================+
    |pure_nccl      |   |OK |Required|``pure_nccl`` is recommended when     |
    |               |   |   |(>= v2) |NCCL2 is available in the environment.|
    +---------------+---+---+--------+--------------------------------------+
    |flat           |   |OK |        |N/A                                   |
    +---------------+---+---+--------+--------------------------------------+
    |naive          |OK |OK |        |Testing on CPU mode                   |
    +---------------+---+---+--------+--------------------------------------+

    pure_nccl communicator supports multiple data types, FP32 and FP16,
    in gradient exchange. The communication data type is determined based on
    `chainer.global_config.dtype` and `allreduce_grad_dtype`.
    When `allreduce_grad_dtype` is the default value `None`,
    FP32 is used when `chainer.global_config.dtype` is `numpy.float32` and
    FP16 otherwise.
    `allreduce_grad_dtype` parameter,
    which is either `numpy.float16` or `numpy.float32`,
    overwrites the `chainer.global_config.dtype`.

    The table blow summarizes the data type selection in gradient exchange.

    +---------------------+--------------------------------------------+
    |                     |              allreduce_grad_dtype          |
    +---------------------+---------+------------------+---------------+
    | global_config.dtype | None    |   numpy.float16  | numpy.float32 |
    +=====================+=========+==================+===============+
    | chainer.mixed16     | FP16    |   FP16           | FP32          |
    +---------------------+---------+------------------+---------------+
    | numpy.float16       | FP16    |   FP16           | FP32          |
    +---------------------+---------+------------------+---------------+
    | numpy.float32       | FP32    |   FP16           | FP32          |
    +---------------------+---------+------------------+---------------+

    Other communicators, namely ``flat`` and ``naive``, support only
    float32 communication, no matter what the model is. This is due to
    MPI's limited support of float16.

    Args:
        communicator_name: The name of communicator (``naive``, ``flat``,
          or ``pure_nccl``)
        mpi_comm: MPI4py communicator
        allreduce_grad_dtype: Data type of gradient used in All-Reduce.
          If ``None``, the dtype of a model is used.

    Returns:
        ChainerMN communicator that implements methods defined in
        :class:`chainermn.CommunicatorBase`

    """

    if mpi_comm is None:
        try:
            import mpi4py.MPI
        except ImportError as e:
            raise ImportError(str(e) + ': '
                              'ChainerMN requires mpi4py for '
                              'distributed training. '
                              'Please read the Chainer official document '
                              'and setup MPI and mpi4py.')
        mpi_comm = mpi4py.MPI.COMM_WORLD

    allreduce_grad_dtype, batched_copy = argument.parse_kwargs(
        kwargs, ('allreduce_grad_dtype', None), ('batched_copy', True))
    argument.assert_kwargs_empty(kwargs)

    if 'batched_copy' in kwargs:
        warnings.warn("The option 'batched_copy' is enabled by default "
                      "it is deprecated, and will be removed in next version",
                      DeprecationWarning)

    if communicator_name != 'pure_nccl' and allreduce_grad_dtype is not None:
        raise ValueError(
            'allreduce_grad_dtype is only available '
            'at \'pure_nccl\' communicator.')

    comm = None

    if communicator_name == 'naive':
        from chainermn.communicators.naive_communicator \
            import NaiveCommunicator
        comm = NaiveCommunicator(mpi_comm=mpi_comm)

    elif communicator_name == 'flat':
        from chainermn.communicators.flat_communicator \
            import FlatCommunicator
        comm = FlatCommunicator(mpi_comm=mpi_comm)

    elif communicator_name == 'non_cuda_aware':
        from chainermn.communicators.non_cuda_aware_communicator \
            import NonCudaAwareCommunicator
        comm = NonCudaAwareCommunicator(mpi_comm=mpi_comm)

    elif communicator_name == 'pure_nccl':
        from chainermn.communicators.pure_nccl_communicator \
            import PureNcclCommunicator
        comm = PureNcclCommunicator(mpi_comm=mpi_comm)
        comm.set_config('allreduce_grad_dtype', allreduce_grad_dtype)

    elif communicator_name == 'dummy':
        from chainermn.communicators.dummy_communicator \
            import DummyCommunicator
        comm = DummyCommunicator(mpi_comm=mpi_comm)

    else:
        raise ValueError(
            'Unrecognized communicator: "{}"'.format(communicator_name))

    # As all currently supported communicators are all ancestor of
    # MpiCommunicator, it is fine calling here for all descendants
    comm.set_config('batched_copy', batched_copy)
    return comm



from onnx_chainer.functions.activation import convert_ClippedReLU  # NOQA
from onnx_chainer.functions.activation import convert_ELU  # NOQA
from onnx_chainer.functions.activation import convert_HardSigmoid  # NOQA
from onnx_chainer.functions.activation import convert_LeakyReLU  # NOQA
from onnx_chainer.functions.activation import convert_LogSoftmax  # NOQA
from onnx_chainer.functions.activation import convert_PReLUFunction  # NOQA
from onnx_chainer.functions.activation import convert_ReLU  # NOQA
from onnx_chainer.functions.activation import convert_Selu  # NOQA
from onnx_chainer.functions.activation import convert_Sigmoid  # NOQA
from onnx_chainer.functions.activation import convert_Softmax  # NOQA
from onnx_chainer.functions.activation import convert_Softplus  # NOQA
from onnx_chainer.functions.activation import convert_Tanh  # NOQA

from onnx_chainer.functions.array import convert_Cast  # NOQA
from onnx_chainer.functions.array import convert_Concat  # NOQA
from onnx_chainer.functions.array import convert_Copy  # NOQA
from onnx_chainer.functions.array import convert_Depth2Space  # NOQA
from onnx_chainer.functions.array import convert_Dstack  # NOQA
from onnx_chainer.functions.array import convert_ExpandDims  # NOQA
from onnx_chainer.functions.array import convert_GetItem  # NOQA
from onnx_chainer.functions.array import convert_Hstack  # NOQA
from onnx_chainer.functions.array import convert_Moveaxis  # NOQA
from onnx_chainer.functions.array import convert_Pad  # NOQA
from onnx_chainer.functions.array import convert_Repeat  # NOQA
from onnx_chainer.functions.array import convert_Reshape  # NOQA
from onnx_chainer.functions.array import convert_ResizeImages  # NOQA
from onnx_chainer.functions.array import convert_Separate  # NOQA
from onnx_chainer.functions.array import convert_Shape  # NOQA
from onnx_chainer.functions.array import convert_Space2Depth  # NOQA
from onnx_chainer.functions.array import convert_SplitAxis  # NOQA
from onnx_chainer.functions.array import convert_Squeeze  # NOQA
from onnx_chainer.functions.array import convert_Stack  # NOQA
from onnx_chainer.functions.array import convert_Swapaxes  # NOQA
from onnx_chainer.functions.array import convert_Tile  # NOQA
from onnx_chainer.functions.array import convert_Transpose  # NOQA
from onnx_chainer.functions.array import convert_Vstack  # NOQA
from onnx_chainer.functions.array import convert_Where  # NOQA

from onnx_chainer.functions.connection import convert_Convolution2DFunction  # NOQA
from onnx_chainer.functions.connection import convert_ConvolutionND  # NOQA
from onnx_chainer.functions.connection import convert_Deconvolution2DFunction  # NOQA
from onnx_chainer.functions.connection import convert_DeconvolutionND  # NOQA
from onnx_chainer.functions.connection import convert_EmbedIDFunction  # NOQA
from onnx_chainer.functions.connection import convert_LinearFunction  # NOQA

from onnx_chainer.functions.loss import convert_SoftmaxCrossEntropy  # NOQA

from onnx_chainer.functions.math import convert_Absolute  # NOQA
from onnx_chainer.functions.math import convert_Add  # NOQA
from onnx_chainer.functions.math import convert_AddConstant  # NOQA
from onnx_chainer.functions.math import convert_Arccos  # NOQA
from onnx_chainer.functions.math import convert_Arcsin  # NOQA
from onnx_chainer.functions.math import convert_Arctan  # NOQA
from onnx_chainer.functions.math import convert_ArgMax  # NOQA
from onnx_chainer.functions.math import convert_ArgMin  # NOQA
from onnx_chainer.functions.math import convert_BroadcastTo  # NOQA
from onnx_chainer.functions.math import convert_Clip  # NOQA
from onnx_chainer.functions.math import convert_Cos  # NOQA
from onnx_chainer.functions.math import convert_Cosh  # NOQA
from onnx_chainer.functions.math import convert_Div  # NOQA
from onnx_chainer.functions.math import convert_DivFromConstant  # NOQA
from onnx_chainer.functions.math import convert_Exp  # NOQA
from onnx_chainer.functions.math import convert_Identity  # NOQA
from onnx_chainer.functions.math import convert_LinearInterpolate  # NOQA
from onnx_chainer.functions.math import convert_Log  # NOQA
from onnx_chainer.functions.math import convert_LogSumExp  # NOQA
from onnx_chainer.functions.math import convert_MatMul  # NOQA
from onnx_chainer.functions.math import convert_Max  # NOQA
from onnx_chainer.functions.math import convert_Maximum  # NOQA
from onnx_chainer.functions.math import convert_Mean  # NOQA
from onnx_chainer.functions.math import convert_Min  # NOQA
from onnx_chainer.functions.math import convert_Minimum  # NOQA
from onnx_chainer.functions.math import convert_Mul  # NOQA
from onnx_chainer.functions.math import convert_MulConstant  # NOQA
from onnx_chainer.functions.math import convert_Neg  # NOQA
from onnx_chainer.functions.math import convert_PowConstVar  # NOQA
from onnx_chainer.functions.math import convert_PowVarConst  # NOQA
from onnx_chainer.functions.math import convert_PowVarVar  # NOQA
from onnx_chainer.functions.math import convert_Prod  # NOQA
from onnx_chainer.functions.math import convert_RsqrtGPU  # NOQA
from onnx_chainer.functions.math import convert_Sin  # NOQA
from onnx_chainer.functions.math import convert_Sinh  # NOQA
from onnx_chainer.functions.math import convert_Sqrt  # NOQA
from onnx_chainer.functions.math import convert_Square  # NOQA
from onnx_chainer.functions.math import convert_Sub  # NOQA
from onnx_chainer.functions.math import convert_SubFromConstant  # NOQA
from onnx_chainer.functions.math import convert_Sum  # NOQA
from onnx_chainer.functions.math import convert_Tan  # NOQA

from onnx_chainer.functions.noise import convert_Dropout  # NOQA

from onnx_chainer.functions.normalization import convert_BatchNormalization  # NOQA
from onnx_chainer.functions.normalization import convert_FixedBatchNormalization  # NOQA
from onnx_chainer.functions.normalization import convert_GroupNormalization  # NOQA
from onnx_chainer.functions.normalization import convert_LocalResponseNormalization  # NOQA
from onnx_chainer.functions.normalization import convert_NormalizeL2  # NOQA

from onnx_chainer.functions.pooling import convert_AveragePooling2D  # NOQA
from onnx_chainer.functions.pooling import convert_AveragePoolingND  # NOQA
from onnx_chainer.functions.pooling import convert_MaxPooling2D  # NOQA
from onnx_chainer.functions.pooling import convert_MaxPoolingND  # NOQA
from onnx_chainer.functions.pooling import convert_ROIPooling2D  # NOQA
from onnx_chainer.functions.pooling import convert_Unpooling2D  # NOQA



from __future__ import absolute_import
import abc
import sys
import threading
import typing as tp  # NOQA
import warnings

import numpy

import chainer
from chainer import backend
from chainer.backends import _cpu
from chainer.backends import cuda
from chainer.backends import intel64
from chainer import types  # NOQA
from chainer import utils
import chainerx


_thread_local = threading.local()

# Used in _ToDeviceVisitor to detect GPU-to-GPU (cupy-to-cupy) device transfer.
# It is usually `None`.
# Assign `False` to enable GPU-to-GPU detection. If detected, `True` will be
# assigned. `None` should be assigned again after retrieving the result.
_thread_local.flag_gpu_to_gpu = None


class DeviceResident(utils.enable_final(meta_base=abc.ABCMeta)):

    """A base class of objects with multi-device hierarchy."""

    _device = _cpu.CpuDevice()

    def __init__(self):
        # Store overridden to_device family method names.
        self._overridden_to_methods = tuple([
            m for m in ('to_cpu', 'to_gpu', 'to_intel64')
            if _is_to_device_method_overridden(self, m)])

    def device_resident_accept(self, visitor):
        """Applies the visitor to all the device objects in this instance.

        Args:
            visitor(~chainer.device_resident.DeviceResidentsVisitor): Visitor.

        This method should be overridden if the concrete class has custom
        sub-hierarchy of device resident objects.
        """
        visitor.visit_device_resident(self)

    @property
    def device(self):
        """:class:`~chainer.backend.Device` instance."""
        return self._device

    @property
    def xp(self) -> types.Xp:
        """Array module corresponding to the device.

        Depending on the device in which this object resides, this property
        returns :mod:`numpy`, :mod:`cupy` or :mod:`chainerx`.

        """
        device = self.device
        if device is None:
            return None
        return device.xp

    @utils.final(action=DeprecationWarning)
    def to_cpu(self) -> 'DeviceResident':
        """Copies parameter variables and persistent values to CPU.

         .. deprecated:: v7.0.0
            Use :meth:`to_device` instead.

        This method does not handle non-registered attributes. If some of such
        attributes must be copied to CPU, the link implementation should
        override :meth:`~DeviceResident.device_resident_accept` to do so.

        Returns: self

        """
        visitor = _ToDeviceVisitor(
            backend.CpuDevice(),
            entry_method_info=('to_cpu', {}),
            starting_device_resident=self)
        self.__to_device(visitor)
        return self

    @utils.final(action=DeprecationWarning)
    def to_gpu(
            self,
            device: tp.Optional[types.CudaDeviceSpec] = None,
    ) -> 'DeviceResident':
        """Copies parameter variables and persistent values to GPU.

         .. deprecated:: v7.0.0
            Use :meth:`to_device` instead.

        This method does not handle non-registered attributes. If some of such
        attributes must be copied to GPU, the link implementation must
        override :meth:`~DeviceResident.device_resident_accept` to do so.

        .. warning::

            This method does not transfer the parameters if they are already on
            GPU. Use ``to_device`` to perform inter-GPU transfer.

        Args:
            device: Target device specifier. If omitted, the current device is
                used.

        Returns: self

        """
        cuda.check_cuda_available()
        cuda_device = cuda._get_device_or_current(device)
        device = chainer.backends.cuda.GpuDevice(cuda_device)
        visitor = _ToDeviceVisitor(
            device,
            entry_method_info=('to_gpu', {'device': device.device}),
            skip_between_cupy_devices=True,
            starting_device_resident=self)
        self.__to_device(visitor)
        return self

    @utils.final(action=DeprecationWarning)
    def to_intel64(self) -> 'DeviceResident':
        """Copies parameter variables and persistent values to CPU.

         .. deprecated:: v7.0.0
            Use :meth:`to_device` instead.

        """
        intel64.check_ideep_available()
        visitor = _ToDeviceVisitor(
            chainer.get_device(intel64.Intel64Device()),
            entry_method_info=('to_intel64', {}),
            starting_device_resident=self)
        self.__to_device(visitor)
        return self

    @utils.final
    def to_chx(self):
        """Converts parameter variables and persistent values to ChainerX \
without any copy.

        This method does not handle non-registered attributes. If some of such
        attributes must be copied to ChainerX, the link implementation must
        override this method to do so.

        Returns: self
        """
        if not chainerx.is_available():
            raise RuntimeError('ChainerX is not available.')

        if self.xp is chainerx:
            return self

        self.device_resident_accept(_ToChxVisitor())
        return self

    @utils.final
    def from_chx(self):
        """Converts parameter variables and persistent values from ChainerX \
to NumPy/CuPy devices without any copy."""
        if self._device.xp is chainerx:
            self._device = self._device.fallback_device

        self.device_resident_accept(_FromChxVisitor())
        return self

    def __to_device(self, to_device_visitor):
        self.device_resident_accept(to_device_visitor)

    @utils.final
    def to_device(
            self,
            device: types.DeviceSpec
    ) -> 'DeviceResident':
        """Copies parameter variables and persistent values to the specified \
device.

        This method does not handle non-registered attributes. If some of such
        attributes must be copied to the device, the link implementation must
        override this method to do so.

        Args:
            device: Target device specifier. See
                :func:`~chainer.get_device` for available values.

        Returns: self

        """
        device = chainer.get_device(device)
        self.__to_device(_ToDeviceVisitor(device))
        return self


def _is_to_device_method_overridden(device_resident, method_name):
    # Returns whether the specified to_device family method is overridden.
    to_method = getattr(device_resident, method_name, None).__func__
    to_method_orig = getattr(DeviceResident, method_name)
    if sys.version_info < (3,):
        to_method_orig = to_method_orig.__func__
    if to_method is not to_method_orig:
        return True  # overridden
    return False


class DeviceResidentsVisitor(object):

    """Base class of visitors that visits device resident objects recursively.

    ..  seealso::
        :class:`chainer.DeviceResident`
    """

    def visit_device_resident(self, device_resident):
        """Processes a :class:`~chainer.DeviceResident` instance."""
        raise NotImplementedError()

    def visit_array(self, arr):
        """Processes an array and returns a new one.

        If the visitor does not create a new array, it can simply return the
        original array.
        """
        raise NotImplementedError()

    def visit_variable(self, param):
        """Processes a :class:`~chainer.Variable` or a \
:class:`~chainer.Parameter`."""
        raise NotImplementedError()


class _ToDeviceVisitor(DeviceResidentsVisitor):
    # A visitor that implements recursive to_device().
    # For backward compatibility, if any of to_cpu/to_gpu/to_intel64 are
    # overridden on a device resident, this visitor calls it instead of
    # `visit_device_resident`. That's true even if `to_device` was originally
    # called.

    def __init__(
            self, device, entry_method_info=None,
            skip_between_cupy_devices=False,
            starting_device_resident=None):

        assert isinstance(device, chainer.backend.Device)

        # `entry_method_info` is for backward compatibility workaround for
        # overridden methods.
        # It indicates which method originally causes this visitor.
        # If it is any of the to_??? method names, descendant resident's
        # respective method will be called if it's overridden
        # (instead of `device_resident_accept`).
        if entry_method_info is not None:
            device_names = {
                'to_cpu': '@numpy',
                'to_gpu': '@cupy:N',
                'to_intel64': '@intel64',
            }
            assert len(entry_method_info) == 2
            method = entry_method_info[0]
            assert method in device_names
            warnings.warn(
                '{} is deprecated. '
                'Please use to_device(\'{}\') instead.'.format(
                    method, device_names[method]),
                DeprecationWarning)

        # starting_device_resident is also for backward compatibility
        # workaround for overridden methods.
        # It is a DeviceResident if to_xxx methods were initially called
        # on this visitor. This is used to avoid infinite accept-visit loop
        # that would occur by calling to_xxx methods.
        assert (starting_device_resident is None
                or isinstance(starting_device_resident, DeviceResident))

        self._device = device
        self._entry_method_info = entry_method_info
        self._skip_between_cupy_devices = skip_between_cupy_devices
        self._starting_device_resident = starting_device_resident

    def visit_device_resident(self, device_resident):
        device_resident._device = self._device

        # Backward compatibility workaround for overridden methods
        if device_resident._overridden_to_methods:
            # Skip this device resident, if the visitor was initially triggered
            # from it.
            if device_resident is self._starting_device_resident:
                return

            if self._entry_method_info is not None:
                # Deprecated method is being called: e.g. to_cpu and to_gpu.
                method_name, kwargs = self._entry_method_info
            else:
                # to_device is being called
                method_name, kwargs = (
                    self._device_to_method_name_and_kwargs(self._device))
            if method_name in device_resident._overridden_to_methods:
                to_method = getattr(device_resident, method_name)
                to_method(**kwargs)
                return

    def _device_to_method_name_and_kwargs(self, device):
        # Converts a device instance to the corresponding combination of
        # to_??? method name and kwargs.

        # chainerx
        if device.xp is chainerx:
            return None, {}
        # cupy
        if device.xp is cuda.cupy:
            return 'to_gpu', {'device': device.device.id}
        # numpy
        assert device.xp is numpy
        if isinstance(device, _cpu.CpuDevice):
            return 'to_cpu', {}
        # intel64
        assert isinstance(device, intel64.Intel64Device)
        return 'to_intel64', {}

    def visit_array(self, arr):
        assert isinstance(arr, chainer.get_array_types())
        device = backend.get_device_from_array(arr)
        if self._skip_visiting(device):
            self._warn_to_gpu(device, self._device)
            return arr
        return self._device.send(arr)

    def visit_variable(self, param):
        assert isinstance(param, chainer.Variable)
        device = param.device
        if self._skip_visiting(device):
            self._warn_to_gpu(device, self._device)
            return
        param.to_device(self._device)

    def _skip_visiting(self, obj_device):
        return (
            self._skip_between_cupy_devices
            and isinstance(self._device, backend.GpuDevice)
            and isinstance(obj_device, backend.GpuDevice))

    @staticmethod
    def _warn_to_gpu(src_device, dst_device):
        src_id = src_device.device.id
        dst_id = dst_device.device.id
        if src_id != dst_id:
            if _thread_local.flag_gpu_to_gpu is None:
                warnings.warn('''\
You are trying to transfer a DeviceResident to GPU-{dst} which is already on \
GPU-{src}.
`DeviceResident.to_gpu` does nothing if the DeviceResident is already on GPU.

You can use `DeviceResident.to_device()` method to perform inter-GPU transfer.
'''.format(dst=dst_id, src=src_id), RuntimeWarning)
            else:
                assert isinstance(_thread_local.flag_gpu_to_gpu, bool)
                _thread_local.flag_gpu_to_gpu = True


class _ToChxVisitor(DeviceResidentsVisitor):
    # A visitor that recursively calls to_chx().

    def visit_device_resident(self, device_resident):
        device_resident._device = backend.ChainerxDevice.from_fallback_device(
            device_resident._device)

    def visit_array(self, arr):
        assert isinstance(arr, chainer.get_array_types())
        return backend.to_chx(arr)

    def visit_variable(self, param):
        assert isinstance(param, chainer.Variable)
        param.to_chx()


class _FromChxVisitor(DeviceResidentsVisitor):
    # A visitor that recursively calls from_chx().

    def visit_device_resident(self, device_resident):
        if isinstance(device_resident._device, backend.ChainerxDevice):
            device_resident._device = device_resident._device.fallback_device

    def visit_array(self, arr):
        assert isinstance(arr, chainer.get_array_types())
        return backend.from_chx(arr)

    def visit_variable(self, param):
        assert isinstance(param, chainer.Variable)
        param.from_chx()



import os
import warnings

import chainer
from chainer.serializers import npz
from chainer.training import extension
from chainer.training.extensions import snapshot_writers
from chainer.utils import argument


def _find_snapshot_files(fmt, path):
    '''Only prefix and suffix match

    TODO(kuenishi): currently clean format string such as
    "snapshot{.iteration}.npz" can only be parsed, but tricky (or
    invalid) formats like "snapshot{{.iteration}}.npz" are hard to
    detect and to properly show errors, just ignored or fails so far.

    Args:
        fmt (str): format string to match with file names of
            existing snapshots, where prefix and suffix are
            only examined. Also, files' staleness is judged
            by timestamps. The default is metime.
        path (str): a directory path to search for snapshot files.

    Returns:
        A sorted list of pair of ``mtime, filename``, whose file
        name that matched the format ``fmt`` directly under ``path``.

    '''
    prefix = fmt.split('{')[0]
    suffix = fmt.split('}')[-1]

    matched_files = (file for file in os.listdir(path)
                     if file.startswith(prefix) and file.endswith(suffix))

    def _prepend_mtime(f):
        t = os.stat(os.path.join(path, f)).st_mtime
        return (t, f)

    return sorted(_prepend_mtime(file) for file in matched_files)


def _find_latest_snapshot(fmt, path):
    """Finds the latest snapshots in a directory

    Args:
        fmt (str): format string to match with file names of
            existing snapshots, where prefix and suffix are
            only examined. Also, files' staleness is judged
            by timestamps. The default is metime.
        path (str): a directory path to search for snapshot files.

    Returns:
        Latest snapshot file, in terms of a file that has newest
        ``mtime`` that matches format ``fmt`` directly under
        ``path``. If no such file found, it returns ``None``.

    """
    snapshot_files = _find_snapshot_files(fmt, path)

    if len(snapshot_files) > 0:
        _, filename = snapshot_files[-1]
        return filename

    return None


def _find_stale_snapshots(fmt, path, n_retains, **kwargs):
    """Finds stale snapshots in a directory, retaining several files

    Args:
        fmt (str): format string to match with file names of
            existing snapshots, where prefix and suffix are
            only examined. Also, files' staleness is judged
            by timestamps. The default is metime.
        path (str): a directory path to search for snapshot files.
        n_retains (int): Number of snapshot files to retain
            through the cleanup. Must be a positive integer for any cleanup to
            take place.
        num_retain (int): Same as n_retains (deprecated).

    Returns:
        Generator that yields stale files that matches format
        ``fmt`` directly under ``path`` and with older ``mtime``,
        excluding newest ``n_retains`` files.

    """
    if 'num_retain' in kwargs:
        warnings.warn(
            'Argument `num_retain` is deprecated. '
            'Please use `n_retains` instead',
            DeprecationWarning)
        n_retains = kwargs['num_retain']

    snapshot_files = _find_snapshot_files(fmt, path)
    num_remove = len(snapshot_files) - n_retains
    if num_remove > 0:
        for _, filename in snapshot_files[:num_remove]:
            yield filename
    return


def snapshot_object(target, filename, savefun=None, **kwargs):
    """snapshot_object(target, filename, savefun=None, \
*, condition=None, writer=None, snapshot_on_error=False, \
num_retain=-1, autoload=False)

    Returns a trainer extension to take snapshots of a given object.

    This extension serializes the given object and saves it to the output
    directory.

    This extension is called once per epoch by default. To take a
    snapshot at a different interval, a trigger object specifying the
    required interval can be passed along with this extension
    to the `extend()` method of the trainer.

    The default priority is -100, which is lower than that of most
    built-in extensions.

    Args:
        target: Object to serialize.
        filename (str): Name of the file into which the object is serialized.
            It can be a format string, where the trainer object is passed to
            the :meth:`str.format` method. For example,
            ``'snapshot_{.updater.iteration}'`` is converted to
            ``'snapshot_10000'`` at the 10,000th iteration.
        savefun: Function to save the object. It takes two arguments: the
            output file path and the object to serialize.
        condition: Condition object. It must be a callable object that returns
            boolean without any arguments. If it returns ``True``, the snapshot
            will be done.
            If not, it will be skipped. The default is a function that always
            returns ``True``.
        writer: Writer object.
            It must be a callable object.
            See below for the list of built-in writers.
            If ``savefun`` is other than ``None``, this argument must be
            ``None``. In that case, a
            :class:`~chainer.training.extensions.snapshot_writers.SimpleWriter`
            object instantiated with specified ``savefun`` argument will be
            used.
        snapshot_on_error (bool): Whether to take a snapshot in case trainer
            loop has been failed.
        n_retains (int): Number of snapshot files to retain
            through the cleanup. Must be a positive integer for any cleanup to
            take place. Automatic deletion of old snapshots only works when the
            filename is string.
        num_retain (int): Same as n_retains (deprecated).
        autoload (bool): With this enabled, the extension automatically
            finds the latest snapshot and loads the data to the target.
            Automatic loading only works when the filename is a string.

    Returns:
        Snapshot extension object.

    .. seealso::

        - :meth:`chainer.training.extensions.snapshot`
    """
    if 'num_retain' in kwargs:
        warnings.warn(
            'Argument `num_retain` is deprecated. '
            'Please use `n_retains` instead',
            DeprecationWarning)
        kwargs['n_retains'] = kwargs['num_retain']
        kwargs.pop('num_retain')

    return snapshot(target=target, filename=filename, savefun=savefun,
                    **kwargs)


def snapshot(savefun=None,
             filename='snapshot_iter_{.updater.iteration}', **kwargs):
    """snapshot(savefun=None, filename='snapshot_iter_{.updater.iteration}', \
*, target=None, condition=None, writer=None, snapshot_on_error=False, \
num_retain=-1, autoload=False)

    Returns a trainer extension to take snapshots of the trainer.

    This extension serializes the trainer object and saves it to the output
    directory. It is used to support resuming the training loop from the saved
    state.

    This extension is called once per epoch by default. To take a
    snapshot at a different interval, a trigger object specifying the
    required interval can be passed along with this extension
    to the `extend()` method of the trainer.

    The default priority is -100, which is lower than that of most
    built-in extensions.

    .. note::
       This extension first writes the serialized object to a temporary file
       and then rename it to the target file name. Thus, if the program stops
       right before the renaming, the temporary file might be left in the
       output directory.

    Args:
        savefun: Function to save the trainer. It takes two arguments: the
            output file path and the trainer object.
            It is :meth:`chainer.serializers.save_npz` by default.
            If ``writer`` is specified, this argument must be ``None``.
        filename (str): Name of the file into which the trainer is serialized.
            It can be a format string, where the trainer object is passed to
            the :meth:`str.format` method.
        target: Object to serialize. If it is not specified, it will
            be the trainer object.
        condition: Condition object. It must be a callable object that returns
            boolean without any arguments. If it returns ``True``, the snapshot
            will be done.
            If not, it will be skipped. The default is a function that always
            returns ``True``.
        writer: Writer object.
            It must be a callable object.
            See below for the list of built-in writers.
            If ``savefun`` is other than ``None``, this argument must be
            ``None``. In that case, a
            :class:`~chainer.training.extensions.snapshot_writers.SimpleWriter`
            object instantiated with specified ``savefun`` argument will be
            used.
        snapshot_on_error (bool): Whether to take a snapshot in case trainer
            loop has been failed.
        n_retains (int): Number of snapshot files to retain
            through the cleanup. Must be a positive integer for any cleanup to
            take place. Automatic deletion of old snapshots only works when the
            filename is string.
        num_retain (int): Same as n_retains (deprecated).
        autoload (bool): With this enabled, the extension
            automatically finds the latest snapshot and loads the data
            to the target.  Automatic loading only works when the
            filename is a string. It is assumed that snapshots are generated
            by :func:`chainer.serializers.save_npz` .

    Returns:
        Snapshot extension object.

    .. testcode::
       :hide:

       from chainer import training
       class Model(chainer.Link):
           def __call__(self, x):
               return x
       train_iter = chainer.iterators.SerialIterator([], 1)
       optimizer = optimizers.SGD().setup(Model())
       updater = training.updaters.StandardUpdater(
           train_iter, optimizer, device=0)
       trainer = training.Trainer(updater)

    .. admonition:: Using asynchronous writers

        By specifying ``writer`` argument, writing operations can be made
        asynchronous, hiding I/O overhead of snapshots.

        >>> from chainer.training import extensions
        >>> writer = extensions.snapshot_writers.ProcessWriter()
        >>> trainer.extend(extensions.snapshot(writer=writer), \
trigger=(1, 'epoch'))

        To change the format, such as npz or hdf5, you can pass a saving
        function as ``savefun`` argument of the writer.

        >>> from chainer.training import extensions
        >>> from chainer import serializers
        >>> writer = extensions.snapshot_writers.ProcessWriter(
        ...     savefun=serializers.save_npz)
        >>> trainer.extend(extensions.snapshot(writer=writer), \
trigger=(1, 'epoch'))

    This is the list of built-in snapshot writers.

        - :class:`chainer.training.extensions.snapshot_writers.SimpleWriter`
        - :class:`chainer.training.extensions.snapshot_writers.ThreadWriter`
        - :class:`chainer.training.extensions.snapshot_writers.ProcessWriter`
        - :class:`chainer.training.extensions.snapshot_writers.\
ThreadQueueWriter`
        - :class:`chainer.training.extensions.snapshot_writers.\
ProcessQueueWriter`

    .. seealso::

        - :meth:`chainer.training.extensions.snapshot_object`
    """
    if 'num_retain' in kwargs:
        warnings.warn(
            'Argument `num_retain` is deprecated. '
            'Please use `n_retains` instead',
            DeprecationWarning)
        kwargs['n_retains'] = kwargs['num_retain']
        kwargs.pop('num_retain')

    target, condition, writer, snapshot_on_error, n_retains,\
        autoload = argument.parse_kwargs(
            kwargs,
            ('target', None), ('condition', None), ('writer', None),
            ('snapshot_on_error', False), ('n_retains', -1),
            ('autoload', False))
    argument.assert_kwargs_empty(kwargs)

    if savefun is not None and writer is not None:
        raise TypeError(
            'savefun and writer arguments cannot be specified together.')

    if writer is None:
        if savefun is None:
            savefun = npz.save_npz
        writer = snapshot_writers.SimpleWriter(savefun=savefun)

    return _Snapshot(
        target=target, condition=condition, writer=writer, filename=filename,
        snapshot_on_error=snapshot_on_error, n_retains=n_retains,
        autoload=autoload)


def _always_true():
    return True


class _Snapshot(extension.Extension):
    """Trainer extension to take snapshots.

    This extension serializes the given object and saves it to the output
    directory.

    This extension is called once per epoch by default. To take a
    snapshot at a different interval, a trigger object specifying the
    required interval can be passed along with this extension
    to the `extend()` method of the trainer.

    The default priority is -100, which is lower than that of most
    built-in extensions.
    """
    trigger = 1, 'epoch'
    priority = -100

    def __init__(
            self, target=None, condition=None, writer=None,
            filename='snapshot_iter_{.updater.iteration}',
            snapshot_on_error=False, n_retains=-1, autoload=False, **kwargs):
        if condition is None:
            condition = _always_true
        if writer is None:
            writer = snapshot_writers.SimpleWriter()
        if 'num_retain' in kwargs:
            warnings.warn(
                'Argument `num_retain` is deprecated. '
                'Please use `n_retains` instead',
                DeprecationWarning)
            n_retains = kwargs['num_retain']

        self._target = target
        self.filename = filename
        self.condition = condition
        self.writer = writer
        self._snapshot_on_error = snapshot_on_error
        self.n_retains = n_retains
        self.autoload = autoload

    def initialize(self, trainer):
        target = trainer if self._target is None else self._target
        outdir = trainer.out
        if self.autoload:
            # If ``autoload`` is on, this code scans the ``outdir``
            # for potential snapshot files by matching the file names
            # from ``filename`` format, picks up the latest one in
            # terms of mtime, and tries to load it it the target or
            # trainer.
            filename = _find_latest_snapshot(self.filename, outdir)
            if filename is None:
                if chainer.is_debug():
                    print('No snapshot file that matches {} was found'
                          .format(self.filename))
            else:
                snapshot_file = os.path.join(outdir, filename)
                # As described above (at ``autoload`` option),
                # snapshot files to be autoloaded must be saved by
                # ``save_npz`` . In order to support general format,
                # we nned to first reconstruct the design of savefun
                # and loadfun.
                npz.load_npz(snapshot_file, target)
                if chainer.is_debug():
                    print('Snapshot loaded from', snapshot_file)

        if (hasattr(self.writer, '_add_cleanup_hook')
                and self.n_retains > 0
                and isinstance(self.filename, str)):
            # This block sets a method to automatic cleanup of stale
            # snapshots, when ``n_retains`` argument is positive
            # number. When the given snapshot writer is Chainer's
            # built-in writer, a cleanup method that is to be
            # triggered right after creation of new snapshot file, is
            # injected here.
            def _cleanup():
                files = _find_stale_snapshots(self.filename, outdir,
                                              self.n_retains)
                for file in files:
                    os.remove(os.path.join(outdir, file))

            self.writer._add_cleanup_hook(_cleanup)

    def on_error(self, trainer, exc, tb):
        super(_Snapshot, self).on_error(trainer, exc, tb)
        if self._snapshot_on_error:
            self._make_snapshot(trainer)

    def __call__(self, trainer):
        if self.condition():
            self._make_snapshot(trainer)

    def _make_snapshot(self, trainer):
        target = trainer if self._target is None else self._target
        serialized_target = npz.serialize(target)
        filename = self.filename
        if callable(filename):
            filename = filename(trainer)
        else:
            filename = filename.format(trainer)
        outdir = trainer.out
        self.writer(filename, outdir, serialized_target)

    def finalize(self):
        if hasattr(self.writer, 'finalize'):
            self.writer.finalize()



import inspect

import chainer


class WrappedFunctionNode(chainer.FunctionNode):
    """Wrap the target function and operate as ``FunctionNode``

    Arguments:
        name (str): name of the function node
        func (func): the target function
        args (list): args for the function
        kwargs (dict): kwargs for the function
        arg_vars (list): list of `chainer.Variable`s in `args` and `kwargs`
        attributes (list): parameters to be set node's attributes
    """

    def __init__(self, name, func, args, kwargs, arg_vars, attributes=None):
        self.custom_function_node_name = name
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.arg_vars = arg_vars
        self.internal_results = None

        if attributes is not None:
            for k, v in attributes.items():
                setattr(self, k, v)

    def forward(self, xs):
        assert len(xs) == len(self.arg_vars)
        self.xs = xs
        results = self.func(*self.args, **self.kwargs)
        if isinstance(results, (tuple, list)):
            dummy_results = tuple(_unwrap_var(ret) for ret in results)
            if all([_is_var(ret) for ret in results]):
                self.internal_results = tuple(results)
        elif isinstance(results, dict):
            dummy_results = tuple(_unwrap_var(ret) for ret in results.values())
            if all([_is_var(ret) for ret in results.values()]):
                self.internal_results = tuple(results.values())
        else:
            dummy_results = _unwrap_var(results)
            dummy_results = dummy_results,
            if _is_var(results):
                self.internal_results = results,
        if not chainer.is_arrays_compatible(dummy_results):
            raise ValueError(
                'returned values from the function wrapped by \'as_funcnode\' '
                'must consist only array, function name: {}'.format(self.name))
        return dummy_results

    def backward(self, target_input_indexes, grad_outputs):
        if self.internal_results is None:
            raise ValueError(
                'the target function does not support backward, propagation '
                'is failed')
        grad_inputs = chainer.grad(self.internal_results, self.arg_vars,
                                   grad_outputs=grad_outputs)
        assert len(self.arg_vars) == len(grad_inputs)
        return tuple(grad_input if i in target_input_indexes else None
                     for i, grad_input in enumerate(grad_inputs))


def fake_as_funcnode(alt_func, name, rename_attributes=None):
    """The target function fakes FunctionNode

    The target function is replaced to the alternative function to connect
    variable node by acting function node. ``alt_func`` must satisfy the
    following restrictions.

    1. Inputs includes one or more ``chainer.Variable`` to trace variables.
    2. Output consists nothing but ``ndarray`` or ``chainer.Variable``

    Even if ``alt_func`` returns ``ndarray``, the value forced to be converted
    to ``chainer.Variable``. A caller of the target function have to care
    both cases, returning ``ndarray`` and ``chainer.Variable``.

    When ``alt_func`` returns ``list`` of variable, the wrapped function will
    also returns multiple variables as ``tuple``. However ``dict`` cannot
    be return, the wrapped function breaks down the returned values as
    ``tuple`` of values, keys will be ignored.

    Arguments of ``alt_func`` except for ``chainer.Variable`` are set as
    function attributes. Attribute names are set ``argN`` (N is index
    number) or keyword on default.

    Example:

       >>> def func(x, a, b, c=1, d=2): pass
       >>> # x is variable
       >>> func = onnx_chainer.replace_func.fake_as_funcnode(
       ...     func, 'CustomNode',
       ...     rename_attributes=[(1, 'value'), ('c', 'y')])

    Then ``func`` will be operated as a function node named "CustomNode", and
    ``'value'``, ``'b'``, ``'y'``, ``'d'`` are set as function's attributes.
    See tests/test_replace_func.py more details.

    Args:
        alt_func (func): actual called function. There are some constrains, see
            the above documentation.
        name (str): function name. This name is used for what ONNX operator
            to be assigned.
        rename_attributes (list or tuple): rename attribute name, set list
            of ``tuple(index_of_args, new_name)`` or
            ``tuple(kwargs_name, new_name)``

    Returns:
        func: wrapped function, called on exporting.
    """

    def _wrapper(*args, **kwargs):
        inputs = []
        attributes = {}
        rename_attr_dict = {}
        if rename_attributes is not None:
            rename_attr_dict = {attr[0]: attr[1] for attr in rename_attributes}

        # resolve default value for kwargs
        arg_spec = inspect.signature(alt_func)
        bound = arg_spec.bind(*args, **kwargs)
        bound.apply_defaults()
        # default values are set on `bound.arguments`, but cannot get them
        # from `bound.kwargs`
        for i, (k, v) in enumerate(bound.arguments.items()):
            if i < len(args):
                continue
            kwargs[k] = v

        def set_attr(key, value):
            default_name = key if isinstance(key, str) else 'arg{}'.format(key)
            attributes[rename_attr_dict.get(key, default_name)] = value

        def expand_args(args_iter):
            for i, a in args_iter:
                if _is_var(a):
                    inputs.append(a)
                elif isinstance(a, (tuple, list)):
                    # all elements are variable -> add flatten them to inputs
                    # all elements are not variable -> add them to attributes
                    # mixed variable and other type value -> error
                    flatten_arg = _flatten(a)
                    var_or_not = map(_is_var, flatten_arg)
                    if all(var_or_not):
                        inputs.extend(flatten_arg)
                    elif not any(var_or_not):
                        set_attr(i, a)
                    else:
                        raise ValueError(
                            'arguments mixed variable and other type are not '
                            'supported')
                else:
                    set_attr(i, a)

        expand_args(enumerate(args))
        expand_args(kwargs.items())
        if not inputs:
            raise ValueError(
                'arguments of the function wrapped by \'as_funcnode\' '
                'must include at least one chainer.Variable, function name: '
                '{}'.format(name))

        wrapped = WrappedFunctionNode(
            name, alt_func, args, kwargs, inputs, attributes=attributes)
        ret = wrapped.apply(inputs)
        if len(ret) > 1:
            return ret
        return ret[0]

    chainer.utils.experimental('as_funcnode')
    return _wrapper


def as_funcnode(name, rename_attributes=None):
    """The target function fakes FunctionNode

    The target function is overwrapped to connect variable node by acting
    function node. Expected to be used as decorator. More detail, see
    ``fake_as_funcnode`` documentation.

    Example:

       >>> @onnx_chainer.replace_func.as_funcnode(
       ...     'CustomNode', rename_attributes=[(1, 'value'), ('c', 'y')])
       ... def func(x, a, b, c=1, d=2): pass

    Args:
        name (str): function name. This name is used for what ONNX operator
            to be assigned.
        rename_attributes (list or tuple): rename attribute name, set list
            of ``tuple(index_of_args, new_name)`` or
            ``tuple(kwargs_name, new_name)``
    """
    def _wrapper(fn):
        return fake_as_funcnode(fn, name, rename_attributes=rename_attributes)

    return _wrapper


def _unwrap_var(var):
    return var.array if _is_var(var) else var


def _is_var(array):
    # alias for type checking
    return isinstance(array, chainer.Variable)


def _is_array(v):
    return not isinstance(v, (list, tuple))


def _flatten(xs):
    if _is_array(xs):
        return [xs]

    o = []
    for x in xs:
        if _is_array(x):
            o.append(x)
        else:
            o.extend(_flatten(x))
    return o



import pytest

pytest.register_assert_rewrite('chainerx.testing.array')
pytest.register_assert_rewrite('chainerx.testing.helper')

from chainerx._testing import _DeviceBuffer  # NOQA
from chainerx._testing import _fromnumpy  # NOQA

from chainerx.testing import array  # NOQA
from chainerx.testing import helper  # NOQA

from chainerx.testing.array import assert_allclose  # NOQA
from chainerx.testing.array import assert_allclose_ex  # NOQA
from chainerx.testing.array import assert_array_equal  # NOQA
from chainerx.testing.array import assert_array_equal_ex  # NOQA
from chainerx.testing.dtypes import all_dtypes  # NOQA
from chainerx.testing.dtypes import float_dtypes  # NOQA
from chainerx.testing.dtypes import integral_dtypes  # NOQA
from chainerx.testing.dtypes import signed_integral_dtypes  # NOQA
from chainerx.testing.dtypes import nonfloat_dtypes  # NOQA
from chainerx.testing.dtypes import numeric_dtypes  # NOQA
from chainerx.testing.dtypes import parametrize_dtype_specifier  # NOQA
from chainerx.testing.dtypes import signed_dtypes  # NOQA
from chainerx.testing.dtypes import unsigned_dtypes  # NOQA
from chainerx.testing.helper import ignore  # NOQA
from chainerx.testing.helper import numpy_chainerx_allclose  # NOQA
from chainerx.testing.helper import numpy_chainerx_array_equal  # NOQA



import warnings

from chainer.utils import conv
import numpy as np

from onnx_chainer.functions.opset_version import support
from onnx_chainer import onnx_helper


@support((1, 7))
def convert_AveragePooling2D(
        func, opset_version, input_names, output_names, context):
    pad = [func.ph, func.pw]
    stride = [func.sy, func.sx]
    ksize = [func.kh, func.kw]
    if func.cover_all:
        # Supports cover_all by setting extra padding
        # NOTE: onnxruntime may not run when "k <= p + s - 1".
        pad.extend([p + s - 1 for p, s in zip(pad, stride)])
    else:
        pad = pad * 2

    if opset_version == 1:
        raise ValueError(
            'AveragePooling2D is not compatible with ONNX\'s AveragePool-1. '
            'Use operation set version >= 7.')
    elif opset_version == 7:
        return onnx_helper.make_node(
            'AveragePool', input_names, output_names,
            kernel_shape=ksize,
            pads=pad,
            strides=stride,
            count_include_pad=1,
        ),


@support((1, 7))
def convert_AveragePoolingND(
        func, opset_version, input_names, output_names, context):
    pad = list(func.pad[:])
    if func.cover_all:
        # Supports cover_all by setting extra padding
        # NOTE: onnxruntime may not run when "k <= p + s - 1".
        pad.extend([p + s - 1 for p, s in zip(pad, func.stride)])
    else:
        pad = pad * 2

    if opset_version == 1:
        raise ValueError(
            'AveragePoolingND is not compatible with ONNX\'s AveragePool-1. '
            'Use operation set version >= 7.')
    elif opset_version == 7:
        return onnx_helper.make_node(
            'AveragePool', input_names, output_names,
            kernel_shape=func.ksize,
            pads=pad,
            strides=func.stride,
            count_include_pad=1,
        ),


@support((1, 8))
def convert_MaxPooling2D(
        func, opset_version, input_names, output_names, context):
    pad = [func.ph, func.pw]
    stride = [func.sy, func.sx]
    ksize = [func.kh, func.kw]
    if func.cover_all:
        # Supports cover_all by setting extra padding
        # NOTE: onnxruntime may not run when "k <= p + s - 1".
        pad.extend([p + s - 1 for p, s in zip(pad, stride)])
    else:
        pad = pad * 2

    if opset_version == 1:
        return onnx_helper.make_node(
            'MaxPool', input_names, output_names,
            kernel_shape=ksize,
            pads=pad,
            strides=stride
        ),
    elif opset_version == 8:
        return onnx_helper.make_node(
            'MaxPool', input_names, output_names,
            kernel_shape=ksize,
            pads=pad,
            strides=stride,
            storage_order=0,  # row major
        ),


@support((1, 8))
def convert_MaxPoolingND(
        func, opset_version, input_names, output_names, context):
    pad = list(func.pad[:])
    if func.cover_all:
        # Supports cover_all by setting extra padding
        # NOTE: onnxruntime may not run when "k <= p + s - 1".
        pad.extend([p + s - 1 for p, s in zip(pad, func.stride)])
    else:
        pad = pad * 2

    if opset_version == 1:
        return onnx_helper.make_node(
            'MaxPool', input_names, output_names,
            kernel_shape=func.ksize,
            pads=pad,
            strides=func.stride
        ),
    elif opset_version == 8:
        return onnx_helper.make_node(
            'MaxPool', input_names, output_names,
            kernel_shape=func.ksize,
            pads=pad,
            strides=func.stride,
            storage_order=0,  # row major
        ),


def convert_ROIPooling2D(
        func, opset_version, input_names, output_names, context):
    warnings.warn(
        'It\'s possible that output does not match with Chainer, please check '
        'each runtime\'s implementation. For example, when input x has '
        'negative values, some runtimes set max(output, 0) unlike Chainer.',
        UserWarning)
    return onnx_helper.make_node(
        'MaxRoiPool', input_names, output_names,
        pooled_shape=[func.outh, func.outw],
        spatial_scale=func.spatial_scale,
    ),


@support((7, 9, 10))
def convert_Unpooling2D(
        func, opset_version, input_names, output_names, context):
    pad = [func.ph, func.pw]
    stride = [func.sy, func.sx]
    ksize = [func.kh, func.kw]
    outsize = [func.outh, func.outw]
    # TODO(hamaji): These could be implemented by `Slice` and `Pad`.
    if func.cover_all:
        raise RuntimeError('ONNX-chainer does not support `cover_all=True` '
                           'for Unpooling2D')
    h, w = func.inputs[0].shape[2:]
    expected_outsize = [
        conv.get_deconv_outsize(
            h, func.kh, func.sy, func.ph, cover_all=func.cover_all),
        conv.get_deconv_outsize(
            w, func.kh, func.sy, func.ph, cover_all=func.cover_all)
    ]
    if outsize != expected_outsize:
        raise RuntimeError('ONNX-chainer does not support `outsize!=None` '
                           'for Unpooling2D: expected={} actual={}'.format(
                               expected_outsize, outsize))
    if pad != [0, 0]:
        raise RuntimeError('ONNX-chainer does not support `pad!=0` '
                           'for Unpooling2D')
    # This one would require an extra 1x1 MaxPool.
    if stride != ksize:
        raise RuntimeError('ONNX-chainer does not support `stride!=ksize` '
                           'for Unpooling2D: stride={} ksize={}'.format(
                               stride, ksize))
    scales = [1.0, 1.0, float(func.kh), float(func.kw)]
    if opset_version == 7:
        return onnx_helper.make_node('Upsample', input_names, output_names,
                                     scales=scales),
    if opset_version in [9, 10]:
        scales_name = context.add_const(
            np.array(scales, dtype=np.float32), 'scales')
        input_names.append(scales_name)
        op = 'Upsample' if opset_version == 9 else 'Resize'
        return onnx_helper.make_node(op, input_names, output_names),



import numpy

import chainer
from chainer import backend
from chainer.backends import cuda
from chainer import distribution
import chainer.distributions.utils
from chainer.functions.activation import sigmoid
from chainer.functions.array import where
from chainer.functions.math import exponential
from chainer.functions.math import logarithm_1p
from chainer.functions.math import sum
from chainer import utils
from chainer.utils import cache


class BernoulliLogProb(chainer.function_node.FunctionNode):

    def __init__(self, binary_check=False):
        super(BernoulliLogProb, self).__init__()
        self.binary_check = binary_check

    def forward(self, inputs):
        logit, x = inputs
        self.retain_inputs((0, 1))
        xp = backend.get_array_module(x)
        y = logit * (x - 1) - xp.log(xp.exp(-logit) + 1)
        y = utils.force_array(y)

        # extreme logit
        logit_isinf = xp.isinf(logit)
        self.logit_ispinf = xp.bitwise_and(logit_isinf, logit > 0)
        self.logit_isminf = xp.bitwise_and(logit_isinf, logit <= 0)
        with numpy.errstate(divide='ignore', invalid='raise'):
            y = xp.where(self.logit_ispinf, xp.log(x), y)
            y = xp.where(self.logit_isminf, xp.log(1 - x), y)

        if self.binary_check:
            self.invalid = utils.force_array(xp.bitwise_and(x != 0, x != 1))
            y[self.invalid] = -xp.inf

        return utils.force_array(y, logit.dtype),

    def backward(self, indexes, grad_outputs):
        gy, = grad_outputs
        logit, x = self.get_retained_inputs()
        xp = backend.get_array_module(x)
        dlogit = x - 1. / (1. + exponential.exp(-logit))

        # extreme logit
        nan = xp.array(xp.nan).astype(dlogit.dtype)
        logit_isinf = xp.bitwise_or(self.logit_ispinf, self.logit_isminf)
        dlogit = where.where(logit_isinf, nan, dlogit)
        if self.binary_check:
            dlogit = where.where(self.invalid, nan, dlogit)

        return sum.sum_to(gy * dlogit, logit.shape), None


def _bernoulli_log_prob(logit, x, binary_check=False):
    y, = BernoulliLogProb(binary_check).apply((logit, x))
    return y


class Bernoulli(distribution.Distribution):

    """Bernoulli Distribution.

    The probability mass function of the distribution is expressed as

    .. math::
        P(x = 1; p) = p \\\\
        P(x = 0; p) = 1 - p

    Args:
        p(:class:`~chainer.Variable` or :ref:`ndarray`): Parameter of
            distribution representing :math:`p`. Either `p` or `logit` (not
            both) must have a value.
        logit(:class:`~chainer.Variable` or :ref:`ndarray`) Parameter of
            distribution representing :math:`\\log\\{p/(1-p)\\}`. Either `p`
            or `logit` (not both) must have a value.

    """

    def __init__(self, p=None, logit=None, binary_check=False):
        super(Bernoulli, self).__init__()
        if not (p is None) ^ (logit is None):
            raise ValueError(
                'Either `p` or `logit` (not both) must have a value.')

        self.__p = p
        self.__logit = logit
        self.binary_check = binary_check

    @cache.cached_property
    def p(self):
        if self.__p is not None:
            return chainer.as_variable(self.__p)
        else:
            return sigmoid.sigmoid(self.logit)

    @cache.cached_property
    def logit(self):
        if self.__logit is not None:
            return chainer.as_variable(self.__logit)
        else:
            return exponential.log(self.p) - logarithm_1p.log1p(-self.p)

    @property
    def batch_shape(self):
        return self.p.shape

    @property
    def entropy(self):
        p = self.p
        q = p.dtype.type(1.) - p
        return (- chainer.distributions.utils._modified_xlogx(p)
                - chainer.distributions.utils._modified_xlogx(q))

    @property
    def event_shape(self):
        return ()

    @property
    def _is_gpu(self):
        return isinstance(self.p.array, cuda.ndarray)

    def log_prob(self, x):
        return _bernoulli_log_prob(self.logit, x, self.binary_check)

    @cache.cached_property
    def mean(self):
        return self.p

    @property
    def params(self):
        return {'logit': self.logit}

    def prob(self, x):
        x = chainer.as_variable(x)
        prob = x * self.p + (1 - x) * (1 - self.p)
        if self.binary_check:
            if self._is_gpu:
                valid = cuda.cupy.bitwise_or(x.array == 0, x.array == 1)
            else:
                valid = numpy.bitwise_or(x.array == 0, x.array == 1)
            prob *= valid
        return prob

    def sample_n(self, n):
        if self._is_gpu:
            eps = cuda.cupy.random.binomial(
                1, self.p.array, size=(n,)+self.p.shape)
        else:
            eps = numpy.random.binomial(
                1, self.p.array, size=(n,)+self.p.shape)
        return chainer.Variable(eps)

    @cache.cached_property
    def stddev(self):
        return self.variance ** 0.5

    @property
    def support(self):
        return '{0, 1}'

    @cache.cached_property
    def variance(self):
        return self.p * (1 - self.p)


@distribution.register_kl(Bernoulli, Bernoulli)
def _kl_bernoulli_bernoulli(dist1, dist2):
    return (
        (dist1.logit - dist2.logit) * (dist1.p - 1.)
        - exponential.log(exponential.exp(-dist1.logit) + 1)
        + exponential.log(exponential.exp(-dist2.logit) + 1))



import numpy

import chainer
from chainer import backend
from chainer import initializer
from chainer import types  # NOQA


class Identity(initializer.Initializer):

    """Initializes array with the identity matrix.

    It initializes the given array with the constant
    multiple of the identity matrix.
    Note that arrays to be passed must be 2D squared matrices.

    Attributes:
        scale (scalar): A constant to be multiplied to identity matrices.

    """

    def __init__(self, scale=1.0, dtype=None):
        self.scale = scale
        super(Identity, self).__init__(dtype)

    def __call__(self, array):
        if self.dtype is not None:
            assert array.dtype == self.dtype
        shape = array.shape
        if len(shape) != 2 or shape[0] != shape[1]:
            raise ValueError('Identity matrix initialization can only be used '
                             'for 2D squared matrices.')

        device = backend.get_device_from_array(array)
        array[...] = device.xp.identity(shape[0]) * self.scale


class _Constant(initializer.Initializer):

    fill_value = None  # type: types.ScalarValue

    def __init__(self, dtype=None):
        if not (isinstance(self.fill_value, chainer.get_array_types())
                or numpy.isscalar(self.fill_value)):
            raise ValueError(
                'fill_value must be either scalar, numpy.ndarray, '
                'cupy.ndarray or chainerx.ndarray.')
        super(_Constant, self).__init__(dtype)

    def __call__(self, array):
        if self.dtype is not None:
            assert array.dtype == self.dtype

        # Calling copy to ensures that the fill_value array
        # is moved to the device where array resides
        if isinstance(self.fill_value, chainer.get_array_types()):
            backend.copyto(
                array, self.fill_value.astype(array.dtype, copy=False))
        else:
            device = backend.get_device_from_array(array)
            array[...] = device.xp.asarray(self.fill_value)


class Constant(_Constant):

    """Initializes array with constant value.

    Attributes:
        ~Constant.fill_value (scalar or :ref:`ndarray`):
            A constant to be assigned to the initialized array.
            Broadcast is allowed on this assignment.
        dtype: Data type specifier.

    """

    def __init__(self, fill_value, dtype=None):
        self.fill_value = fill_value
        super(Constant, self).__init__(dtype)


class Zero(_Constant):
    """Initializes array to all-zero.

    Attributes:
        ~Zero.dtype: Data type specifier.
    """

    fill_value = 0.0


class One(_Constant):
    """Initializes array to all-one.

    Attributes:
        ~One.dtype: Data type specifier.
    """

    fill_value = 1.0


class NaN(_Constant):
    """Initializes array to all-NaN.

    Attributes:
        ~NaN.dtype: Data type specifier.
    """

    fill_value = numpy.nan



#!/usr/bin/env python

import os
import pkg_resources
import sys

from setuptools import setup

import chainerx_build_helper


if sys.version_info[:3] == (3, 5, 0):
    if not int(os.getenv('CHAINER_PYTHON_350_FORCE', '0')):
        msg = """
Chainer does not work with Python 3.5.0.

We strongly recommend to use another version of Python.
If you want to use Chainer with Python 3.5.0 at your own risk,
set CHAINER_PYTHON_350_FORCE environment variable to 1."""
        print(msg)
        sys.exit(1)


requirements = {
    'install': [
        'setuptools',
        # typing==3.7.4 causes error "TypeError: Instance and class checks can
        # only be used with @runtime_checkable protocols" only with Python 2.
        # https://github.com/chainer/chainer/pull/7562
        'typing' + ('<=3.6.6' if sys.version_info[0] <= 2 else ''),
        'typing_extensions' + ('<=3.6.6' if sys.version_info[0] <= 2 else ''),
        'filelock',
        'numpy>=1.9.0',
        'protobuf>=3.0.0',
        'six>=1.9.0',
    ],
    'stylecheck': [
        'autopep8>=1.4.1,<1.5',
        'flake8>=3.7,<3.8',
        'pycodestyle>=2.5,<2.6',
    ],
    'test': [
        'pytest<4.2.0',  # 4.2.0 is slow collecting tests and times out on CI.
        'attrs<19.2.0',  # pytest 4.1.1 does not run with attrs==19.2.0
        'mock',
    ],
    'doctest': [
        'sphinx==1.8.2',
        'matplotlib',
        'theano',
    ],
    'docs': [
        'sphinx==1.8.2',
        'sphinx_rtd_theme',
        'onnx<1.6.0',
        'packaging',
    ],
    'appveyor': [
        '-r test',
        # pytest-timeout>=1.3.0 requires pytest>=3.6.
        # TODO(niboshi): Consider upgrading pytest to >=3.6
        'pytest-timeout<1.3.0',
    ],
    'jenkins': [
        '-r test',
        # pytest-timeout>=1.3.0 requires pytest>=3.6.
        # TODO(niboshi): Consider upgrading pytest to >=3.6
        'pytest-timeout<1.3.0',
        'pytest-cov',
        'nose',
        'coveralls',
        'codecov',
    ],
}


def reduce_requirements(key):
    # Resolve recursive requirements notation (-r)
    reqs = requirements[key]
    resolved_reqs = []
    for req in reqs:
        if req.startswith('-r'):
            depend_key = req[2:].lstrip()
            reduce_requirements(depend_key)
            resolved_reqs += requirements[depend_key]
        else:
            resolved_reqs.append(req)
    requirements[key] = resolved_reqs


for k in requirements.keys():
    reduce_requirements(k)


extras_require = {k: v for k, v in requirements.items() if k != 'install'}
setup_requires = []
install_requires = requirements['install']
tests_require = requirements['test']


def find_any_distribution(pkgs):
    for pkg in pkgs:
        try:
            return pkg_resources.get_distribution(pkg)
        except pkg_resources.DistributionNotFound:
            pass
    return None


for pkg_name in ('ChainerMN', 'ONNX-Chainer'):
    distribution_name = pkg_name.lower().replace('-', '_')
    found_error = find_any_distribution([distribution_name])
    if found_error is not None:
        msg = """
We detected that {name} is installed in your environment.
{name} has been integrated to Chainer and no separate installation
is necessary. Please uninstall the old {name} in advance.
"""
        print(msg.format(name=pkg_name))
        exit(1)

here = os.path.abspath(os.path.dirname(__file__))
# Get __version__ variable
exec(open(os.path.join(here, 'chainer', '_version.py')).read())


setup_kwargs = dict(
    name='chainer',
    version=__version__,  # NOQA
    description='A flexible framework of neural networks',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Seiya Tokui',
    author_email='tokui@preferred.jp',
    url='https://chainer.org/',
    license='MIT License',
    packages=['chainer',
              'chainer.backends',
              'chainer.dataset',
              'chainer.dataset.tabular',
              'chainer.datasets',
              'chainer.distributions',
              'chainer.exporters',
              'chainer.functions',
              'chainer.functions.activation',
              'chainer.functions.array',
              'chainer.functions.connection',
              'chainer.functions.evaluation',
              'chainer.functions.loss',
              'chainer.functions.math',
              'chainer.functions.noise',
              'chainer.functions.normalization',
              'chainer.functions.pooling',
              'chainer.functions.rnn',
              'chainer.functions.theano',
              'chainer.functions.util',
              'chainer.function_hooks',
              'chainer.iterators',
              'chainer.initializers',
              'chainer.links',
              'chainer.links.activation',
              'chainer.links.caffe',
              'chainer.links.caffe.protobuf3',
              'chainer.links.connection',
              'chainer.links.loss',
              'chainer.links.model',
              'chainer.links.model.vision',
              'chainer.links.normalization',
              'chainer.links.rnn',
              'chainer.links.theano',
              'chainer.link_hooks',
              'chainer.graph_optimizations',
              'chainer.optimizers',
              'chainer.optimizer_hooks',
              'chainer.serializers',
              'chainer.testing',
              'chainer.training',
              'chainer.training.extensions',
              'chainer.training.triggers',
              'chainer.training.updaters',
              'chainer.utils',
              'chainermn',
              'chainermn.communicators',
              'chainermn.datasets',
              'chainermn.extensions',
              'chainermn.functions',
              'chainermn.iterators',
              'chainermn.links',
              'chainermn.testing',
              'onnx_chainer',
              'onnx_chainer.functions',
              'onnx_chainer.testing'],
    package_data={
        'chainer': ['py.typed'],
    },
    zip_safe=False,
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
)


build_chainerx = 0 != int(os.getenv('CHAINER_BUILD_CHAINERX', '0'))
if (os.getenv('READTHEDOCS', None) == 'True'
        and os.getenv('READTHEDOCS_PROJECT', None) == 'chainer'):
    # ChainerX must be built in order to build the docs (on Read the Docs).
    build_chainerx = True

    # Try to prevent Read the Docs build timeouts.
    os.environ['MAKEFLAGS'] = '-j2'


chainerx_build_helper.config_setup_kwargs(setup_kwargs, build_chainerx)


setup(**setup_kwargs)



import numbers
import typing as tp  # NOQA
import typing_extensions as tpe  # NOQA

try:
    from typing import TYPE_CHECKING  # NOQA
except ImportError:
    # typing.TYPE_CHECKING doesn't exist before Python 3.5.2
    TYPE_CHECKING = False

# import chainer modules only for type checkers to avoid circular import
if TYPE_CHECKING:
    from types import ModuleType  # NOQA

    import numpy  # NOQA

    from chainer import backend  # NOQA
    from chainer.backends import cuda, intel64  # NOQA
    from chainer import initializer  # NOQA

    import chainerx  # NOQA


Shape = tp.Tuple[int, ...]


ShapeSpec = tp.Union[int, tp.Sequence[int]]  # Sequence includes Tuple[int, ...] # NOQA


DTypeSpec = tp.Union[tp.Any]  # TODO(okapies): encode numpy.dtype


NdArray = tp.Union[
    'numpy.ndarray',
    'cuda.ndarray',
    # 'intel64.mdarray',
    # TODO(okapies): mdarray is partially incompatible with other ndarrays
    'chainerx.ndarray',
]
"""The ndarray types supported in :func:`chainer.get_array_types`
"""


Xp = tp.Union[tp.Any]  # TODO(okapies): encode numpy/cupy/ideep/chainerx


class AbstractInitializer(tpe.Protocol):
    """Protocol class for Initializer.

    It can be either an :class:`chainer.Initializer` or a callable object
    that takes an ndarray.

    This is only for PEP 544 compliant static type checkers.
    """
    dtype = None  # type: tp.Optional[DTypeSpec]

    def __call__(self, array: NdArray) -> None:
        pass


ScalarValue = tp.Union[
    'numpy.generic',
    bytes,
    str,
    memoryview,
    numbers.Number,
]
"""The scalar types supported in :func:`numpy.isscalar`.
"""


InitializerSpec = tp.Union[AbstractInitializer, ScalarValue, 'numpy.ndarray']


DeviceSpec = tp.Union[
    'backend.Device',
    'chainerx.Device',
    'cuda.Device',
    str,
    tp.Tuple[str, int],
    'ModuleType',  # numpy and intel64 module
    tp.Tuple['ModuleType', int],  # cupy module and device ID
]
"""The device specifier types supported in :func:`chainer.get_device`
"""
# TODO(okapies): Use Xp instead of ModuleType


CudaDeviceSpec = tp.Union['cuda.Device', int, 'numpy.integer']  # NOQA
"""
This type only for the deprecated :func:`chainer.cuda.get_device` API.
Use :class:`~chainer.types.DeviceSpec` instead.
"""



import typing as tp  # NOQA

from chainer import types  # NOQA
from chainer import utils


class Initializer(object):
    """Initializes array.

    It initializes the given array.

    Attributes:
        dtype: Data type specifier. It is for type check in ``__call__``
            function.

    """

    def __init__(self, dtype: tp.Optional[types.DTypeSpec] = None) -> None:
        self.dtype = dtype  # type: types.DTypeSpec

    def __call__(self, array: types.NdArray) -> None:
        """Initializes given array.

        This method destructively changes the value of array.
        The derived class is required to implement this method.
        The algorithms used to make the new values depend on the
        concrete derived classes.

        Args:
            array (:ref:`ndarray`):
                An array to be initialized by this initializer.

        """
        raise NotImplementedError()


# Original code forked from MIT licensed keras project
# https://github.com/fchollet/keras/blob/master/keras/initializations.py

def get_fans(shape):
    if not isinstance(shape, tuple):
        raise ValueError(
            'shape must be tuple. Actual type: {}'.format(type(shape)))

    if len(shape) < 2:
        raise ValueError(
            'shape must be of length >= 2. Actual shape: {}'.format(shape))

    receptive_field_size = utils.size_of_shape(shape[2:])
    fan_in = shape[1] * receptive_field_size
    fan_out = shape[0] * receptive_field_size
    return fan_in, fan_out



import chainer
from chainer.functions.normalization import group_normalization
from chainer import link_hook


class WeightStandardization(link_hook.LinkHook):
    """Weight Standardization (WS) link hook implementation.

    This hook standardizes a weight by *weight statistics*.

    This link hook implements a WS which computes the mean and variance along
    axis "output channels", then normalizes by these statistics.
    WS improves training by reducing the Lipschitz constants of the loss and
    the gradients like batch normalization (BN) but without relying on large
    batch sizes during training. Specifically, the performance of WS with group
    normalization (GN) trained with small-batch is able to match or outperforms
    that of BN trained with large-batch.
    WS is originally proposed for 2D convolution layers followed by mainly GN
    and sometimes BN.
    Note that this hook is able to handle layers such as N-dimensional
    convolutional, linear and embedding layers but there is no guarantee that
    this hook helps training.

    See: Siyuan Qiao et. al., `Weight Standardization
    <https://arxiv.org/abs/1903.10520>`_

    Args:
        eps (float): Numerical stability in standard deviation calculation.
            The default value is 1e-5.
        weight_name (str): Link's weight name to appky this hook. The default
            value is ``'W'``.
        name (str or None): Name of this hook. The default value is
            ``'WeightStandardization'``.
    """

    name = 'WeightStandardization'

    def __init__(self, *, eps=1e-5, weight_name='W', name=None):
        self.eps = eps
        self.weight_name = weight_name
        self._initialized = False
        if name is not None:
            self.name = name

    def __enter__(self):
        raise NotImplementedError(
            'This hook is not supposed to be used as context manager.')

    def __exit__(self):
        raise NotImplementedError

    def added(self, link):
        if not hasattr(link, self.weight_name):
            raise ValueError(
                'Weight \'{}\' does not exist!'.format(self.weight_name))
        if getattr(link, self.weight_name).array is not None:
            self._initialized = True

    def forward_preprocess(self, cb_args):
        # This method normalizes target link's weight by statistics
        link = cb_args.link
        input_variable = cb_args.args[0]
        if not self._initialized:
            if getattr(link, self.weight_name).array is None:
                if input_variable is None:
                    raise ValueError('Input variable does not exist!')
                link._initialize_params(input_variable.shape[1])
        weight = getattr(link, self.weight_name)
        with chainer.using_device(link.device):
            gamma = link.xp.ones(
                (weight.shape[1],), dtype=weight.dtype)
            beta = link.xp.zeros(
                (weight.shape[1],), dtype=weight.dtype)
        # For link.W or equivalents to be chainer.Parameter
        # consistently to users, this hook maintains a reference to
        # the unnormalized weight.
        self.original_weight = weight
        # note: `normalized_weight` is ~chainer.Variable
        normalized_weight = group_normalization.group_normalization(
            weight, groups=1, gamma=gamma, beta=beta, eps=self.eps)
        setattr(link, self.weight_name, normalized_weight)

    def forward_postprocess(self, cb_args):
        # Here, the computational graph is already created,
        # we can reset link.W or equivalents to be Parameter.
        link = cb_args.link
        setattr(link, self.weight_name, self.original_weight)



# import classes and functions
from chainer.training.extensions._snapshot import snapshot  # NOQA
from chainer.training.extensions._snapshot import snapshot_object  # NOQA
from chainer.training.extensions.computational_graph import DumpGraph  # NOQA
from chainer.training.extensions.evaluator import Evaluator  # NOQA
from chainer.training.extensions.exponential_shift import ExponentialShift  # NOQA
from chainer.training.extensions.fail_on_nonnumber import FailOnNonNumber  # NOQA
from chainer.training.extensions.inverse_shift import InverseShift  # NOQA
from chainer.training.extensions.linear_shift import LinearShift  # NOQA
from chainer.training.extensions.log_report import LogReport  # NOQA
from chainer.training.extensions.micro_average import MicroAverage  # NOQA
from chainer.training.extensions.multistep_shift import MultistepShift  # NOQA
from chainer.training.extensions.parameter_statistics import ParameterStatistics  # NOQA
from chainer.training.extensions.plot_report import PlotReport  # NOQA
from chainer.training.extensions.polynomial_shift import PolynomialShift  # NOQA
from chainer.training.extensions.print_report import PrintReport  # NOQA
from chainer.training.extensions.progress_bar import ProgressBar  # NOQA
from chainer.training.extensions.step_shift import StepShift  # NOQA
from chainer.training.extensions.value_observation import observe_lr  # NOQA
from chainer.training.extensions.value_observation import observe_value  # NOQA
from chainer.training.extensions.variable_statistics_plot import VariableStatisticsPlot  # NOQA
from chainer.training.extensions.variable_unchain import unchain_variables  # NOQA
from chainer.training.extensions.warmup_shift import WarmupShift  # NOQA

# Alias
from chainer.training.extensions.computational_graph import DumpGraph as dump_graph  # NOQA



class FunctionConverterParams(object):

    def __init__(
            self, func=None, opset_version=None, input_names=None,
            output_names=None, context=None):
        """Wrapper of converter parameters

        Exporter set this parameters to the target converter's argument.

        >>> def own_converter(params):
        ...     # params is FunctionConverterParams
        ...     # so enable to get each attributes:
        ...     func_name = params.func.__class__.__name__

        Arguments:
            func (~chainer.FunctionNode): Target function.
            opset_version (int): Target opset version.
            input_names (list): List of input names.
            output_names (list): List of ouptut names.
            context (~onnx_chainer.context.Context): Context for Exporting
        """

        self.func = func
        self.opset_version = opset_version
        self.input_names = input_names
        self.output_names = output_names
        self.context = context


class FunctionConverter(object):

    def __init__(self, converter):
        """Wrapper of ONNX-Chainer converter

        Exporter set arguments wrapped by ``FunctionConverterParams``, and
        this class breaks downs to each argument.

        Arguments:
            converter (function): The target converter function.
        """

        self.converter = converter

    def __call__(self, params):
        func = params.func
        opset_version = params.opset_version
        input_names = params.input_names
        output_names = params.output_names
        context = params.context
        return self.converter(
            func, opset_version, input_names, output_names, context)



import chainer
from chainer import backend
from chainer import configuration
import chainer.functions as F
from chainer import link_hook
import chainer.links as L
from chainer import variable
import chainerx


def l2normalize(xp, v, eps):
    """Normalize a vector by its L2 norm.

    Args:
        xp (numpy or cupy):
        v (numpy.ndarray or cupy.ndarray)
        eps (float): Epsilon value for numerical stability.

    Returns:
        :class:`numpy.ndarray` or :class:`cupy.ndarray`

    """
    # TODO(crcrpar): Remove this when chainerx.linalg.norm becomes available.
    if xp is chainerx:
        # NOTE(crcrpar): `chainerx.power` is not available as of 2019/03/27.
        # See https://github.com/chainer/chainer/pull/6522
        norm = chainerx.sqrt(chainerx.sum(v * v))
    else:
        norm = xp.linalg.norm(v)
    return v / (norm + eps)


def update_approximate_vectors(
        weight_matrix, u, n_power_iteration, eps):
    """Update the first left and right singular vectors.

    This function updates the first left singular vector `u` and
    the first right singular vector `v`.

    Args:
        weight_matrix (~chainer.Variable): 2D weight.
        u (numpy.ndarray, cupy.ndarray, or None):
            Vector that approximates the first left singular vector and
            has the shape of (out_size,).
        n_power_iteration (int): Number of iterations to approximate
            the first right and left singular vectors.

    Returns:
        :class:`numpy.ndarray` or `cupy.ndarray`:
            Approximate first left singular vector.
        :class:`numpy.ndarray` or `cupy.ndarray`:
            Approximate first right singular vector.

    """
    weight_matrix = weight_matrix.array
    xp = backend.get_array_module(weight_matrix)
    for _ in range(n_power_iteration):
        v = l2normalize(xp, xp.dot(u, weight_matrix), eps)
        u = l2normalize(xp, xp.dot(weight_matrix, v), eps)
    return u, v


def calculate_max_singular_value(weight_matrix, u, v):
    """Calculate max singular value by power iteration method.

    Args:
        weight_matrix (~chainer.Variable)
        u (numpy.ndarray or cupy.ndarray)
        v (numpy.ndarray or cupy.ndarray)

    Returns:
        ~chainer.Variable: Max singular value via power iteration method.

    """
    sigma = F.matmul(F.matmul(u, weight_matrix), v)
    return sigma


class SpectralNormalization(link_hook.LinkHook):
    """Spectral Normalization link hook implementation.

    This hook normalizes a weight using max singular value and this value
    is computed via power iteration method. Currently, this hook is supposed to
    be added to :class:`chainer.links.Linear`, :class:`chainer.links.EmbedID`,
    :class:`chainer.links.Convolution2D`, :class:`chainer.links.ConvolutionND`,
    :class:`chainer.links.Deconvolution2D`,
    and :class:`chainer.links.DeconvolutionND`. However, you can use this to
    other links like RNNs by specifying ``weight_name``.
    It is highly recommended to add this hook before optimizer setup because
    this hook add a scaling parameter ``gamma`` if ``use_gamma`` is True.
    Otherwise, the registered ``gamma`` will not be updated.

    .. math::

       \\bar{\\mathbf{W}} &=& \\dfrac{\\mathbf{W}}{\\sigma(\\mathbf{W})} \\\\
       \\text{, where} \\ \\sigma(\\mathbf{W}) &:=&
        \\max_{\\mathbf{h}: \\mathbf{h} \\ne 0}
       \\dfrac{\\|\\mathbf{W} \\mathbf{h}\\|_2}{\\|\\mathbf{h}\\|_2}
        = \\max_{\\|\\mathbf{h}\\|_2 \\le 1} \\|\\mathbf{W}\\mathbf{h}\\|_2

    See: T. Miyato et. al., `Spectral Normalization for Generative Adversarial
    Networks <https://arxiv.org/abs/1802.05957>`_

    Args:
        n_power_iteration (int): Number of power iteration.
            The default value is 1.
        eps (float): Numerical stability in norm calculation.
            The default value is 1e-6 for the compatibility with
            mixed precision training. The value used in the author's
            implementation is 1e-12.
        use_gamma (bool): If ``True``, weight scaling parameter gamma which is
            initialized by initial weight's max singular value is introduced.
        factor (float, None): Scaling parameter to divide maximum singular
            value.  The default value is 1.0.
        weight_name (str): Link's weight name to apply this hook. The default
            value is ``'W'``.
        name (str or None): Name of this hook. The default value is
            ``'SpectralNormalization'``.

    Attributes:
        vector_name (str): Name of the approximate first left singular vector
            registered in the target link.
            the target link.
        axis (int): Axis of weight represents the number of output
            feature maps or output units (``out_channels`` and
            ``out_size``, respectively).

    .. admonition:: Example

        There are almost the same but 2 ways to apply spectral normalization
        (SN) hook to links.

        1. Initialize link and SN separately. This makes it easy to handle
        buffer and parameter of links registered by SN hook.

            >>> l = L.Convolution2D(3, 5, 3)
            >>> hook = chainer.link_hooks.SpectralNormalization()
            >>> _ = l.add_hook(hook)
            >>> # Check the shape of the first left singular vector.
            >>> getattr(l, hook.vector_name).shape
            (5,)
            >>> # Delete SN hook from this link.
            >>> l.delete_hook(hook.name)

        2. Initialize both link and SN hook at one time. This makes it easy to
        define your original :class:`~chainer.Chain`.

            >>> # SN hook handles lazy initialization!
            >>> layer = L.Convolution2D(
            ...     5, 3, stride=1, pad=1).add_hook(
            ...         chainer.link_hooks.SpectralNormalization())
    """

    name = 'SpectralNormalization'

    def __init__(self, n_power_iteration=1, eps=1e-6, use_gamma=False,
                 factor=None, weight_name='W', name=None):
        assert n_power_iteration > 0
        self.n_power_iteration = n_power_iteration
        self.eps = eps
        self.use_gamma = use_gamma
        self.factor = factor
        self.weight_name = weight_name
        self.vector_name = weight_name + '_u'
        self._initialized = False
        self.axis = 0
        if name is not None:
            self.name = name

    def __enter__(self):
        raise NotImplementedError(
            'This hook is not supposed to be used as context manager.')

    def __exit__(self):
        raise NotImplementedError

    def added(self, link):
        # Define axis and register ``u`` if the weight is initialized.
        if not hasattr(link, self.weight_name):
            raise ValueError(
                'Weight \'{}\' does not exist!'.format(self.weight_name))
        if isinstance(link, (L.Deconvolution2D, L.DeconvolutionND)):
            self.axis = 1
        if getattr(link, self.weight_name).array is not None:
            self._prepare_parameters(link)

    def deleted(self, link):
        # Remove approximate vector ``u`` and parameter ``gamma` if exists.
        delattr(link, self.vector_name)
        if self.use_gamma:
            del link.gamma

    def forward_preprocess(self, cb_args):
        # This method normalizes target link's weight spectrally
        # using power iteration method
        link = cb_args.link
        input_variable = cb_args.args[0]
        if not self._initialized:
            self._prepare_parameters(link, input_variable)
        weight = getattr(link, self.weight_name)
        # For link.W or equivalents to be chainer.Parameter
        # consistently to users, this hook maintains a reference to
        # the unnormalized weight.
        self.original_weight = weight
        # note: `normalized_weight` is ~chainer.Variable
        normalized_weight = self.normalize_weight(link)
        setattr(link, self.weight_name, normalized_weight)

    def forward_postprocess(self, cb_args):
        # Here, the computational graph is already created,
        # we can reset link.W or equivalents to be Parameter.
        link = cb_args.link
        setattr(link, self.weight_name, self.original_weight)

    def _prepare_parameters(self, link, input_variable=None):
        """Prepare one buffer and one parameter.

        Args:
            link (:class:`~chainer.Link`): Link to normalize spectrally.
            input_variable (:class:`~chainer.Variable`):
                The first minibatch to initialize weight.

        """
        if getattr(link, self.weight_name).array is None:
            if input_variable is not None:
                link._initialize_params(input_variable.shape[1])
        initialW = getattr(link, self.weight_name)
        if initialW.shape[self.axis] == 0:
            raise ValueError(
                'Expect {}.shape[{}] > 0'.format(self.weight_name, self.axis)
            )
        u = link.xp.random.normal(
            size=(initialW.shape[self.axis],)).astype(dtype=initialW.dtype)
        setattr(link, self.vector_name, u)
        link.register_persistent(self.vector_name)
        if self.use_gamma:
            # Initialize the scaling parameter with the max singular value.
            weight_matrix = self.reshape_W(initialW.array)
            # TODO(crcrpar): Remove this when chainerx supports SVD.
            device = link.device
            if device.xp is chainerx:
                fallback_device = device.fallback_device
                weight_matrix_ = fallback_device.send(weight_matrix)
                with chainer.using_device(fallback_device):
                    _, s_, _ = fallback_device.xp.linalg.svd(weight_matrix_)
                s = device.send(s_)
            else:
                _, s, _ = link.xp.linalg.svd(weight_matrix)
            s0 = chainer.utils.force_array(s[0])
            with link.init_scope():
                link.gamma = variable.Parameter(s0)
        self._initialized = True

    def normalize_weight(self, link):
        """Normalize target weight before every single forward computation."""
        weight_name, vector_name = self.weight_name, self.vector_name
        W = getattr(link, weight_name)
        u = getattr(link, vector_name)
        weight_matrix = self.reshape_W(W)
        if not configuration.config.in_recomputing:
            with chainer.using_device(link.device):
                u, v = update_approximate_vectors(
                    weight_matrix, u, self.n_power_iteration, self.eps)
        else:
            v = self.v
        sigma = calculate_max_singular_value(weight_matrix, u, v)
        if self.factor is not None:
            sigma /= self.factor
        if self.use_gamma:
            W = link.gamma * W / sigma
        else:
            W = W / sigma
        if not configuration.config.in_recomputing:
            self.v = v
            with chainer.using_device(link.device):
                if configuration.config.train:
                    if link.xp is chainerx:
                        # TODO(crcrpar): Remove this when
                        # chainerx supports `copyto`.
                        getattr(link, vector_name)[:] = u
                    else:
                        backend.copyto(getattr(link, vector_name), u)
        return W

    def reshape_W(self, W):
        """Reshape & transpose weight into 2D if necessary."""
        if self.axis != 0:
            axes = [self.axis] + [i for i in range(W.ndim) if i != self.axis]
            W = W.transpose(axes)
        if W.ndim == 2:
            return W
        return W.reshape(W.shape[0], -1)



import typing as tp  # NOQA

import numpy

import chainer
from chainer import backend
from chainer.backends import _chainerx  # NOQA

# import class and function
from chainer.initializers.constant import Constant
from chainer.initializers.constant import Identity  # NOQA
from chainer.initializers.constant import NaN  # NOQA
from chainer.initializers.constant import One  # NOQA
from chainer.initializers.constant import Zero  # NOQA
from chainer.initializers.normal import GlorotNormal  # NOQA
from chainer.initializers.normal import HeNormal  # NOQA
from chainer.initializers.normal import LeCunNormal
from chainer.initializers.normal import Normal  # NOQA
from chainer.initializers.orthogonal import Orthogonal  # NOQA
from chainer.initializers.sampling import DownsamplingConvFilter  # NOQA
from chainer.initializers.sampling import UpsamplingDeconvFilter  # NOQA
from chainer.initializers.uniform import GlorotUniform  # NOQA
from chainer.initializers.uniform import HeUniform  # NOQA
from chainer.initializers.uniform import LeCunUniform  # NOQA
from chainer.initializers.uniform import Uniform  # NOQA
from chainer import types  # NOQA


def generate_array(
        initializer: types.AbstractInitializer,
        shape: types.ShapeSpec,
        xp: types.Xp,
        dtype: tp.Optional[types.DTypeSpec] = None,
        device: tp.Optional[types.DeviceSpec] = None
) -> types.NdArray:
    """Return initialized array.

    The algorithms used to make the new values depend on the
    concrete derived classes. If the initializer has the ``dtype`` attribute,
    it is used to construct the array. Otherwise, ``chainer.config.dtype`` is
    used instead. See :ref:`configuration` for the dtype config.

    Args:
        initializer: A callable object that takes :ref:`ndarray` and edits its
            value.
        shape (int or tuple of int): Shape of the initialized array.
        xp (module): :mod:`cupy`, :mod:`numpy`, or :mod:`chainerx`.
        dtype: Dtype specifier. If omitted, ``initializer.dtype`` is used.
        device: Target device specifier. If omitted, the current device is
             used for :mod:`cupy`, and the default device is used for
             :mod:`chainerx`.

    Returns:
        :ref:`ndarray`: An initialized array.

    """
    dtype_attr = getattr(initializer, 'dtype', None)
    if dtype is not None and dtype_attr is not None \
            and numpy.dtype(dtype) != numpy.dtype(dtype_attr):
        raise ValueError(
            'dtype mismatch: {} != {}'.format(dtype, dtype_attr))
    if dtype is None:
        dtype = dtype_attr
    dtype = chainer.get_dtype(dtype)

    if device is None:
        backend_device = backend._guess_device_from_array_module(xp)
    else:
        backend_device = chainer.get_device(device)
        if xp != backend_device.xp:
            raise ValueError('xp and device arguments are inconsistent.')
    with chainer.using_device(backend_device):
        array = xp.empty(shape, dtype=dtype)
        initializer(array)
    return array


def _get_initializer(
        initializer: tp.Optional[types.InitializerSpec]
) -> types.AbstractInitializer:
    if initializer is None:
        return LeCunNormal()
    if (isinstance(initializer, chainer.get_array_types())
            or numpy.isscalar(initializer)):
        return Constant(initializer)

    if not callable(initializer):
        raise TypeError('invalid type of initializer: %s' % type(initializer))
    return initializer


def _check_is_initializer_like(initializer):
    if not (initializer is None
            or isinstance(initializer, chainer.Initializer)
            or callable(initializer)
            or isinstance(initializer, chainer.get_array_types())
            or numpy.isscalar(initializer)):
        raise TypeError(
            'Initializer is of wrong type: {}. Allowed types are Initializer, '
            'ndarray and scalar.'.format(type(initializer)))



import chainerx


def _recover_device(backend_name, device_index):
    # Recovers the device instance.
    # This function is used together with chainerx.Device.__reduce__.
    # TODO(niboshi): Save the context name and lookup the context with it.
    context = chainerx.get_default_context()
    backend = context.get_backend(backend_name)
    device = backend.get_device(device_index)
    return device
